// 64-Switch (Command-Line Pass-Through Target)
// by Hendricks266

#include "64switch.h"

wchar_t* getFileFromPath(wchar_t* path)
{
    wchar_t* file = path;

    file = wcsrchr(file,'/');
    if (file != NULL)
        path = file;
    else
        file = path;

    file = wcsrchr(file,'\\');
    if (file != NULL)
        path = file;
    else
        file = path;

    if (file[0] == '/' || file[0] == '\\')
        file += 1;

    return file;
}

DWORD ErrorOutput(BOOL value, const wchar_t* name)
{
    if (!value)
    {
        LPTSTR errorText = NULL;
        DWORD error = GetLastError();

        // from http://stackoverflow.com/a/455533
        FormatMessage(
           // use system message tables to retrieve error text
           FORMAT_MESSAGE_FROM_SYSTEM
           // allocate buffer on local heap for error text
           |FORMAT_MESSAGE_ALLOCATE_BUFFER
           // Important! will fail otherwise, since we will not 
           // (and CANNOT) pass insertion parameters
           |FORMAT_MESSAGE_IGNORE_INSERTS,  
           NULL,    // unused with FORMAT_MESSAGE_FROM_SYSTEM
           error,
           MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
           (LPTSTR)&errorText,  // output 
           0, // minimum size for output buffer
           NULL);   // arguments

        if ( NULL != errorText )
        {
           fwprintf(stderr,L"%s error:\n%s",name,errorText);

           // release memory allocated by FormatMessage()
           LocalFree(errorText);
           errorText = NULL;
        }

        return error;
    }

    return 0;
}

int main(void)
{
    wchar_t *exename = NULL;
	DWORD lpExitCode = 0;

    {
        wchar_t originalexepath[MAX_PATH];
        GetModuleFileNameW(NULL, originalexepath, MAX_PATH);
        wchar_t *originalexename = getFileFromPath(originalexepath);

        int originalexenamelength = wcslen(originalexename);
        int originalexepathlength = wcslen(originalexepath)-originalexenamelength;

        wchar_t *exenameafterfirstdot = wcschr(originalexename,'.');
        int exenameafterfirstdotlength = 0;
        if (exenameafterfirstdot)
            exenameafterfirstdotlength = wcslen(exenameafterfirstdot);

        wchar_t *exenameextension = wcsstr(originalexename+(originalexenamelength-exenameafterfirstdotlength),L".exe");
        int exenameextensionlength = 0;
        if (exenameextension)
            exenameextensionlength = wcslen(exenameextension);


        int exeprefixlength = originalexenamelength-exenameafterfirstdotlength;
        exename = malloc(sizeof(wchar_t)*(originalexepathlength+originalexenamelength+4));
        wcsncpy(exename,originalexepath,originalexepathlength);
        exename[originalexepathlength]=0;
        wcsncat(exename,originalexename,exeprefixlength);
        exename[originalexepathlength+exeprefixlength]=0;

        {
            const wchar_t *PROCESSOR_ARCHITECTURE = wgetenv(L"PROCESSOR_ARCHITECTURE");
            const wchar_t *PROCESSOR_ARCHITEW6432 = wgetenv(L"PROCESSOR_ARCHITEW6432");
            if ((PROCESSOR_ARCHITECTURE && wcscasecmp(PROCESSOR_ARCHITECTURE,L"amd64") == 0) || (PROCESSOR_ARCHITEW6432 && wcscasecmp(PROCESSOR_ARCHITEW6432,L"amd64") == 0))
                wcscat(exename,L"-64");
            else if ((PROCESSOR_ARCHITECTURE && wcscasecmp(PROCESSOR_ARCHITECTURE,L"x86") == 0) && !PROCESSOR_ARCHITEW6432)
                wcscat(exename,L"-32");
            else
                wcscat(exename,L"-32");
        }

        wcsncat(exename,originalexename+exeprefixlength,exenameafterfirstdotlength-exenameextensionlength);
        wcscat(exename,L".exe");
    }

    {
        SECURITY_ATTRIBUTES secatt;
        secatt.nLength = sizeof(SECURITY_ATTRIBUTES);
        secatt.lpSecurityDescriptor = NULL;
        secatt.bInheritHandle = TRUE;

        STARTUPINFOW lpStartupInfo;
        PROCESS_INFORMATION lpProcessInformation;

        GetStartupInfoW(&lpStartupInfo);

        if (!ErrorOutput(CreateProcessW(exename,GetCommandLineW(),&secatt,&secatt,TRUE,GetPriorityClass(GetCurrentProcess()),NULL,NULL,&lpStartupInfo,&lpProcessInformation),L"CreateProcessWithTokenW"))
        {
            WaitForSingleObject(lpProcessInformation.hProcess, INFINITE);

            ErrorOutput(GetExitCodeProcess(lpProcessInformation.hProcess,&lpExitCode),L"GetExitCodeProcess");

            ErrorOutput(CloseHandle(lpProcessInformation.hProcess),L"CloseHandle (hProcess)");
            ErrorOutput(CloseHandle(lpProcessInformation.hThread),L"CloseHandle (hThread)");
        }
        else
            fwprintf(stderr,L"%s\n",exename);
    }

    if (exename)
        free(exename);

    return lpExitCode;
}
